#include "stdafx.h"
#include "MAC.h"
#include "All.h"
#include "FLACTag.h"
#include "IO.h"

namespace APE
{

#define Swap4Bytes(val) \
((((val) >> 24) & 0x000000FF) | (((val) >> 8) & 0x0000FF00) | \
(((val) << 8) & 0x00FF0000) | (((val) << 24) & 0xFF000000))

/**************************************************************************************************
CFLACTag
**************************************************************************************************/
CFLACTag::CFLACTag(const TCHAR * pFilename)
{
    // initiliaze
    m_nPictureBytes = 0;

    // open the file
    CSmartPtr<APE::CIO> m_spIO;
    m_spIO.Assign(CreateCIO());
    if (m_spIO->Open(pFilename) == ERROR_SUCCESS)
    {
        // read FLAC header
        unsigned int nBytesRead = 4;
        char cFLAC[5] = { 0 };
        m_spIO->Read(cFLAC, 4, &nBytesRead);
        if (strcmp(cFLAC, "fLaC") == 0)
        {
            // loop chunks
            while (true)
            {
                // read the buffer
                unsigned char cBuffer[4] = { 0 };
                if (m_spIO->Read(cBuffer, 4, &nBytesRead) != ERROR_SUCCESS)
                    break;

                // buffer size
                int nBufferBytes = 0;
                {
                    // flip the buffer
                    for (int z = 0; z < 3; z++)
                        nBufferBytes = (nBufferBytes << 8) | (uint32_t)(cBuffer[1 + z]);

                    // get type and last
                    int nLast = static_cast<std::uint8_t>(cBuffer[0] & 0x80);
                    int nType = static_cast<std::uint8_t>(cBuffer[0] & (0x80 - 1));

                    // keep start position of the reader
                    int64 nOriginalPosition = m_spIO->GetPosition();

                    // switch based on the type
                    #define FLAC__METADATA_TYPE_VORBIS_COMMENT 4
                    #define FLAC__METADATA_TYPE_PICTURE 6
                    if (nType == FLAC__METADATA_TYPE_VORBIS_COMMENT)
                    {
                        // tag
                        CSmartPtr<char> spTag;
                        spTag.Assign(new char[nBufferBytes + 1], true);
                        m_spIO->Read(spTag, nBufferBytes, &nBytesRead);

                        // skip vendor
                        int nStartByte = 0;
                        {
                            int nBytes = 0;
                            char* p = &spTag[nStartByte];
                            nBytes = *((int*)p);

                            CSmartPtr<char> spChar(new char [nBytes + 1]);
                            memcpy(spChar, &spTag[nStartByte + 4], nBytes);
                            spChar[nBytes] = 0;

                            nStartByte += nBytes + 4;// +4;
                        }

                        // read fields
                        int nFields = 0;
                        nFields = *((int *) &spTag[nStartByte]);

                        // increment start byte
                        nStartByte += 4;

                        for (int nField = 0; nField < nFields; nField++)
                        {
                            int nBytes = *((int *) &spTag[nStartByte]);

                            CSmartPtr<APE::str_utf8> spChar(new APE::str_utf8[nBytes + 1]);
                            memcpy(spChar, &spTag[nStartByte + 4], nBytes);
                            spChar[nBytes] = 0;

                            CString strTag = CAPECharacterHelper::GetUTF16FromUTF8(spChar.GetPtr());
                            int nEqual = strTag.Find(_T("="), 0);
                            if (nEqual >= 0)
                            {
                                CString strKey = strTag.Left(nEqual);
                                CString strValue = strTag.Mid(nEqual + 1);

                                m_mapFields.SetAt(strKey, strValue);
                            }

                            nStartByte += nBytes + 4;
                        }
                    }
                    else if (nType == FLAC__METADATA_TYPE_PICTURE)
                    {
                        // picture
                        CSmartPtr<char> spPicture;
                        spPicture.Assign(new char [nBufferBytes], true);
                        m_spIO->Read(spPicture, nBufferBytes, &nBytesRead);

                        // picture type
                        int nPictureType = Swap4Bytes(*((int *) &spPicture[0]));

                        // MIME
                        int nMIMELength = Swap4Bytes(*((int *) &spPicture[4]));
                        CSmartPtr<char> spMimeType(new char [nMIMELength + 1], true);
                        memcpy(spMimeType, &spPicture[8], nMIMELength);
                        spMimeType[nMIMELength] = 0;

                        if (_stricmp(spMimeType, "image/jpeg") == 0)
                        {
                            // description
                            int nDescriptionLength = Swap4Bytes(*((int *) &spPicture[8 + nMIMELength]));
                            // we don't read the description because it's not used
                            
                            // width
                            int nWidth = Swap4Bytes(*((int *) &spPicture[12 + nMIMELength + nDescriptionLength]));

                            // height
                            int nHeight = Swap4Bytes(*((int *) &spPicture[16 + nMIMELength + nDescriptionLength]));

                            // depth
                            int nDepth = Swap4Bytes(*((int*) &spPicture[20 + nMIMELength + nDescriptionLength]));

                            // index
                            int nIndex = Swap4Bytes(*((int *) &spPicture[24 + nMIMELength + nDescriptionLength]));

                            // length
                            m_nPictureBytes = Swap4Bytes(*((int *) &spPicture[28 + nMIMELength + nDescriptionLength]));

                            // read data
                            if ((32 + nMIMELength + nDescriptionLength + m_nPictureBytes) <= nBufferBytes)
                            {
                                m_spPicture.Assign(new char [m_nPictureBytes], true);
                                memcpy(m_spPicture, &spPicture[32 + nMIMELength + nDescriptionLength], m_nPictureBytes);
                            }
                        }
                    }

                    // skip the buffer
                    m_spIO->Seek(nOriginalPosition + nBufferBytes, APE::SeekFileBegin);
                }
            }
        }
    }
}

}